@@ -15,7 +15,6 @@ import click_completion
 
 from click_didyoumean import DYMCommandCollection
 
-from .. import environments
 from ..__version__ import __version__
 from .options import (
     CONTEXT_SETTINGS, PipenvGroup, code_option, common_options, deploy_option,
@@ -28,16 +27,20 @@ from .options import (
 # Enable shell completion.
 click_completion.init()
 
+subcommand_context = CONTEXT_SETTINGS.copy()
+subcommand_context.update({
+    "ignore_unknown_options": True,
+    "allow_extra_args": True
+})
+subcommand_context_no_interspersion = subcommand_context.copy()
+subcommand_context_no_interspersion["allow_interspersed_args"] = False
+
 
 @group(cls=PipenvGroup, invoke_without_command=True, context_settings=CONTEXT_SETTINGS)
 @option("--where", is_flag=True, default=False, help="Output project home information.")
 @option("--venv", is_flag=True, default=False, help="Output virtualenv information.")
-@option(
-    "--py", is_flag=True, default=False, help="Output Python interpreter information."
-)
-@option(
-    "--envs", is_flag=True, default=False, help="Output Environment Variable options."
-)
+@option("--py", is_flag=True, default=False, help="Output Python interpreter information.")
+@option("--envs", is_flag=True, default=False, help="Output Environment Variable options.")
 @option("--rm", is_flag=True, default=False, help="Remove the virtualenv.")
 @option("--bare", is_flag=True, default=False, help="Minimal output.")
 @option(
@@ -98,12 +101,12 @@ def cli(
         warn_in_virtualenv,
         do_where,
         project,
-        spinner,
         cleanup_virtualenv,
         ensure_project,
         format_help,
         do_clear,
     )
+    from ..utils import create_spinner
 
     if man:
         if system_which("man"):
@@ -115,6 +118,7 @@ def cli(
             return 1
     if envs:
         echo("The following environment variables can be set, to do various things:\n")
+        from .. import environments
         for key in environments.__dict__:
             if key.startswith("PIPENV"):
                 echo("  - {0}".format(crayons.normal(key, bold=True)))
@@ -161,7 +165,8 @@ def cli(
         # --rm was passed…
         elif rm:
             # Abort if --system (or running in a virtualenv).
-            if environments.PIPENV_USE_SYSTEM:
+            from ..environments import PIPENV_USE_SYSTEM
+            if PIPENV_USE_SYSTEM:
                 echo(
                     crayons.red(
                         "You are attempting to remove a virtualenv that "
@@ -179,7 +184,7 @@ def cli(
                         )
                     )
                 )
-                with spinner():
+                with create_spinner(text="Running..."):
                     # Remove the virtualenv.
                     cleanup_virtualenv(bare=True)
                 return 0
@@ -210,7 +215,7 @@ def cli(
 
 @cli.command(
     short_help="Installs provided packages and adds them to Pipfile, or (if no packages are given), installs all packages from Pipfile.",
-    context_settings=dict(ignore_unknown_options=True, allow_extra_args=True),
+    context_settings=subcommand_context,
 )
 @system_option
 @code_option
@@ -252,8 +257,11 @@ def install(
         ctx.abort()
 
 
-@cli.command(short_help="Un-installs a provided package and removes it from Pipfile.")
-@option("--lock", is_flag=True, default=True, help="Lock afterwards.")
+@cli.command(
+    short_help="Un-installs a provided package and removes it from Pipfile.",
+    context_settings=subcommand_context
+)
+@option("--skip-lock/--lock", is_flag=True, default=False, help="Lock afterwards.")
 @option(
     "--all-dev",
     is_flag=True,
@@ -268,36 +276,39 @@ def install(
 )
 @uninstall_options
 @pass_state
+@pass_context
 def uninstall(
+    ctx,
     state,
-    lock=False,
+    skip_lock=False,
     all_dev=False,
     all=False,
     **kwargs
 ):
     """Un-installs a provided package and removes it from Pipfile."""
     from ..core import do_uninstall
-
     retcode = do_uninstall(
         packages=state.installstate.packages,
         editable_packages=state.installstate.editables,
         three=state.three,
         python=state.python,
         system=state.system,
-        lock=lock,
+        lock=not state.installstate.skip_lock,
         all_dev=all_dev,
         all=all,
         keep_outdated=state.installstate.keep_outdated,
         pypi_mirror=state.pypi_mirror,
+        ctx=ctx
     )
     if retcode:
         sys.exit(retcode)
 
-
-@cli.command(short_help="Generates Pipfile.lock.")
+@cli.command(short_help="Generates Pipfile.lock.", context_settings=CONTEXT_SETTINGS)
 @lock_options
 @pass_state
+@pass_context
 def lock(
+    ctx,
     state,
     **kwargs
 ):
@@ -307,9 +318,14 @@ def lock(
     # Ensure that virtualenv is available.
     ensure_project(three=state.three, python=state.python, pypi_mirror=state.pypi_mirror)
     if state.installstate.requirementstxt:
-        do_init(dev=state.installstate.dev, requirements=state.installstate.requirementstxt,
-                        pypi_mirror=state.pypi_mirror, pre=state.installstate.pre)
+        do_init(
+            dev=state.installstate.dev,
+            requirements=state.installstate.requirementstxt,
+            pypi_mirror=state.pypi_mirror,
+            pre=state.installstate.pre,
+        )
     do_lock(
+        ctx=ctx,
         clear=state.clear,
         pre=state.installstate.pre,
         keep_outdated=state.installstate.keep_outdated,
@@ -319,7 +335,7 @@ def lock(
 
 @cli.command(
     short_help="Spawns a shell within the virtualenv.",
-    context_settings=dict(ignore_unknown_options=True, allow_extra_args=True),
+    context_settings=subcommand_context,
 )
 @option(
     "--fancy",
@@ -378,11 +394,7 @@ def shell(
 @cli.command(
     add_help_option=False,
     short_help="Spawns a command installed into the virtualenv.",
-    context_settings=dict(
-        ignore_unknown_options=True,
-        allow_interspersed_args=False,
-        allow_extra_args=True,
-    ),
+    context_settings=subcommand_context_no_interspersion,
 )
 @common_options
 @argument("command")
@@ -399,7 +411,7 @@ def run(state, command, args):
 
 @cli.command(
     short_help="Checks for security vulnerabilities and against PEP 508 markers provided in Pipfile.",
-    context_settings=dict(ignore_unknown_options=True, allow_extra_args=True),
+    context_settings=subcommand_context
 )
 @option(
     "--unused",
@@ -439,7 +451,7 @@ def check(
     )
 
 
-@cli.command(short_help="Runs lock, then sync.")
+@cli.command(short_help="Runs lock, then sync.", context_settings=CONTEXT_SETTINGS)
 @option("--bare", is_flag=True, default=False, help="Minimal output.")
 @option(
     "--outdated", is_flag=True, default=False, help=u"List out-of-date dependencies."
@@ -516,7 +528,10 @@ def update(
     )
 
 
-@cli.command(short_help=u"Displays currently-installed dependency graph information.")
+@cli.command(
+    short_help=u"Displays currently-installed dependency graph information.",
+    context_settings=CONTEXT_SETTINGS
+)
 @option("--bare", is_flag=True, default=False, help="Minimal output.")
 @option("--json", is_flag=True, default=False, help="Output JSON.")
 @option("--json-tree", is_flag=True, default=False, help="Output JSON in nested tree.")
@@ -528,17 +543,28 @@ def graph(bare=False, json=False, json_tree=False, reverse=False):
     do_graph(bare=bare, json=json, json_tree=json_tree, reverse=reverse)
 
 
-@cli.command(short_help="View a given module in your editor.", name="open")
+@cli.command(
+    short_help="View a given module in your editor.", name="open",
+    context_settings=CONTEXT_SETTINGS
+)
 @common_options
 @argument("module", nargs=1)
 @pass_state
 def run_open(state, module, *args, **kwargs):
-    """View a given module in your editor."""
+    """View a given module in your editor.
+
+    This uses the EDITOR environment variable. You can temporarily override it,
+    for example:
+
+        EDITOR=atom pipenv open requests
+    """
     from ..core import which, ensure_project
 
     # Ensure that virtualenv is available.
-    ensure_project(three=state.three, python=state.python, validate=False,
-                        pypi_mirror=state.pypi_mirror)
+    ensure_project(
+        three=state.three, python=state.python,
+        validate=False, pypi_mirror=state.pypi_mirror,
+    )
     c = delegator.run(
         '{0} -c "import {1}; print({1}.__file__);"'.format(which("python"), module)
     )
@@ -556,7 +582,10 @@ def run_open(state, module, *args, **kwargs):
     return 0
 
 
-@cli.command(short_help="Installs all packages specified in Pipfile.lock.")
+@cli.command(
+    short_help="Installs all packages specified in Pipfile.lock.",
+    context_settings=CONTEXT_SETTINGS
+)
 @option("--bare", is_flag=True, default=False, help="Minimal output.")
 @sync_options
 @pass_state
@@ -589,7 +618,11 @@ def sync(
         ctx.abort()
 
 
-@cli.command(short_help="Uninstalls all packages not specified in Pipfile.lock.")
+@cli.command(
+    short_help="Uninstalls all packages not specified in Pipfile.lock.",
+    context_settings=CONTEXT_SETTINGS
+)
+@option("--bare", is_flag=True, default=False, help="Minimal output.")
 @option("--dry-run", is_flag=True, default=False, help="Just output unneeded packages.")
 @verbose_option
 @three_option
